# chrono - agendas system
# Copyright (C) 2022  Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import decimal

from django.contrib.postgres.fields import JSONField
from django.db import models
from django.template import Context, Template, TemplateSyntaxError
from django.utils.text import slugify
from django.utils.translation import ugettext_lazy as _

from chrono.agendas.models import Agenda, Booking, Subscription
from chrono.utils.misc import generate_slug


class PricingError(Exception):
    def __init__(self, details=None):
        self.details = details or {}
        super().__init__()


class AgendaPricingNotFound(PricingError):
    pass


class CriteriaConditionNotFound(PricingError):
    pass


class PricingDataError(PricingError):
    pass


class PricingDataFormatError(PricingError):
    pass


class PricingEventNotCheckedError(PricingError):
    pass


class PricingBookingNotCheckedError(PricingError):
    pass


class PricingSubscriptionError(PricingError):
    pass


class PricingMultipleBookingError(PricingError):
    pass


class PricingBookingCheckTypeError(PricingError):
    pass


class CriteriaCategory(models.Model):
    label = models.CharField(_('Label'), max_length=150)
    slug = models.SlugField(_('Identifier'), max_length=160, unique=True)

    class Meta:
        ordering = ['label']

    def __str__(self):
        return self.label

    def save(self, *args, **kwargs):
        if not self.slug:
            self.slug = generate_slug(self)
        super().save(*args, **kwargs)

    @property
    def base_slug(self):
        return slugify(self.label)


class Criteria(models.Model):
    category = models.ForeignKey(
        CriteriaCategory, verbose_name=_('Category'), on_delete=models.CASCADE, related_name='criterias'
    )
    label = models.CharField(_('Label'), max_length=150)
    slug = models.SlugField(_('Identifier'), max_length=160)
    condition = models.CharField(_('Condition'), max_length=1000)
    order = models.PositiveIntegerField()

    class Meta:
        ordering = ['order']
        unique_together = ['category', 'slug']

    def __str__(self):
        return self.label

    def save(self, *args, **kwargs):
        if self.order is None:
            max_order = (
                Criteria.objects.filter(category=self.category)
                .aggregate(models.Max('order'))
                .get('order__max')
                or 0
            )
            self.order = max_order + 1
        if not self.slug:
            self.slug = generate_slug(self, category=self.category)
        super().save(*args, **kwargs)

    @property
    def base_slug(self):
        return slugify(self.label)

    def compute_condition(self, context):
        try:
            template = Template('{%% if %s %%}OK{%% endif %%}' % self.condition)
        except TemplateSyntaxError:
            return False
        return template.render(Context(context)) == 'OK'


class Pricing(models.Model):
    label = models.CharField(_('Label'), max_length=150)
    slug = models.SlugField(_('Identifier'), max_length=160, unique=True)
    categories = models.ManyToManyField(
        CriteriaCategory,
        related_name='pricings',
        through='PricingCriteriaCategory',
    )
    criterias = models.ManyToManyField(Criteria)

    class Meta:
        ordering = ['label']

    def __str__(self):
        return self.label

    def save(self, *args, **kwargs):
        if not self.slug:
            self.slug = generate_slug(self)
        super().save(*args, **kwargs)

    @property
    def base_slug(self):
        return slugify(self.label)


class PricingCriteriaCategory(models.Model):
    pricing = models.ForeignKey(Pricing, on_delete=models.CASCADE)
    category = models.ForeignKey(CriteriaCategory, on_delete=models.CASCADE)
    order = models.PositiveIntegerField()

    class Meta:
        ordering = ['order']
        unique_together = ['pricing', 'category']

    def save(self, *args, **kwargs):
        if self.order is None:
            max_order = (
                PricingCriteriaCategory.objects.filter(pricing=self.pricing)
                .aggregate(models.Max('order'))
                .get('order__max')
                or 0
            )
            self.order = max_order + 1
        super().save(*args, **kwargs)


class AgendaPricing(models.Model):
    agenda = models.ForeignKey(Agenda, on_delete=models.CASCADE)
    pricing = models.ForeignKey(Pricing, on_delete=models.CASCADE)
    date_start = models.DateField()
    date_end = models.DateField()
    pricing_data = JSONField(null=True)

    @staticmethod
    def get_pricing_data(event, user_external_id, adult_external_id):
        agenda_pricing = AgendaPricing.get_agenda_pricing(event)
        context = agenda_pricing.get_pricing_context(event, user_external_id, adult_external_id)
        pricing, criterias = agenda_pricing.compute_pricing(context)
        modifier = agenda_pricing.get_booking_modifier(event, user_external_id)
        return agenda_pricing.aggregate_pricing_data(pricing, criterias, context, modifier)

    def aggregate_pricing_data(self, pricing, criterias, context, modifier):
        if modifier['modifier_type'] == 'fixed':
            pricing_amount = modifier['modifier_fixed']
        else:
            pricing_amount = pricing * modifier['modifier_rate'] / 100
        return {
            'pricing': pricing_amount,
            'calculation_details': {
                'pricing': pricing,
                'criterias': criterias,
                'context': context,
            },
            'booking_details': modifier,
        }

    @staticmethod
    def get_agenda_pricing(event):
        agenda = event.agenda
        try:
            return agenda.agendapricing_set.get(
                date_start__lte=event.start_datetime,
                date_end__gt=event.start_datetime,
            )
        except (AgendaPricing.DoesNotExist, AgendaPricing.MultipleObjectsReturned):
            raise AgendaPricingNotFound

    def get_subscription(self, event, user_external_id):
        try:
            return self.agenda.subscriptions.get(
                user_external_id=user_external_id,
                date_start__lte=event.start_datetime,
                date_end__gt=event.start_datetime,
            )
        except (Subscription.DoesNotExist, Subscription.MultipleObjectsReturned):
            raise PricingSubscriptionError

    def get_pricing_context(self, event, user_external_id, adult_external_id):
        # FIXME: compute agenda pricing variables, add event data
        return {
            'qf': 2,
            'domicile': 'commune',
        }

    def compute_pricing(self, context):
        criterias = {}
        categories = []
        # for each category (ordered)
        for category in self.pricing.categories.all().order_by('pricingcriteriacategory__order'):
            criterias[category.slug] = None
            categories.append(category.slug)
            # find the first matching criteria (criterias are ordered)
            for criteria in self.pricing.criterias.filter(category=category):
                condition = criteria.compute_condition(context)
                if condition:
                    criterias[category.slug] = criteria.slug
                    break

        # now search for pricing values matching found criterias
        pricing_data = self.pricing_data
        # for each category (ordered)
        for category in categories:
            criteria = criterias[category]
            if criteria is None:
                raise CriteriaConditionNotFound(details={'category': category})
            if not isinstance(pricing_data, dict):
                raise PricingDataFormatError(
                    details={'category': category, 'pricing': pricing_data, 'wanted': 'dict'}
                )
            key = '%s:%s' % (category, criteria)
            if key not in pricing_data:
                raise PricingDataError(details={'category': category, 'criteria': criteria})
            pricing_data = pricing_data[key]

        try:
            pricing = decimal.Decimal(pricing_data)
        except (decimal.InvalidOperation, ValueError, TypeError):
            raise PricingDataFormatError(details={'pricing': pricing_data, 'wanted': 'decimal'})

        return pricing, criterias

    def get_booking_modifier(self, event, user_external_id):
        # event must be checked
        if event.checked is False:
            raise PricingEventNotCheckedError

        # search for an available subscription
        self.get_subscription(event, user_external_id)

        # search for a booking
        try:
            booking = event.booking_set.get(user_external_id=user_external_id)
        except Booking.DoesNotExist:
            # no booking
            return {
                'status': 'not-booked',
                'modifier_type': 'rate',
                'modifier_rate': 0,
            }
        except Booking.MultipleObjectsReturned:
            raise PricingMultipleBookingError

        # booking cancelled
        if booking.cancellation_datetime is not None:
            return {
                'status': 'cancelled',
                'modifier_type': 'rate',
                'modifier_rate': 0,
            }

        # booking not cancelled, is must be checked
        if booking.user_was_present is None:
            raise PricingBookingNotCheckedError

        status = 'presence' if booking.user_was_present else 'absence'
        # no check_type, default rates
        if booking.user_check_type is None:
            return {
                'status': status,
                'check_type_group': None,
                'check_type': None,
                'modifier_type': 'rate',
                'modifier_rate': 100 if booking.user_was_present else 0,
            }

        check_type = booking.user_check_type
        kind_mapping = {'presence': True, 'absence': False}
        # check_type kind and user_was_present mismatch
        if kind_mapping[check_type.kind] != booking.user_was_present:
            raise PricingBookingCheckTypeError(
                details={
                    'check_type_group': check_type.group.slug,
                    'check_type': check_type.slug,
                    'reason': 'wrong-kind',
                }
            )

        # get pricing modifier
        if check_type.pricing is not None:
            return {
                'status': status,
                'check_type_group': check_type.group.slug,
                'check_type': check_type.slug,
                'modifier_type': 'fixed',
                'modifier_fixed': check_type.pricing,
            }
        if check_type.pricing_rate is not None:
            return {
                'status': status,
                'check_type_group': check_type.group.slug,
                'check_type': check_type.slug,
                'modifier_type': 'rate',
                'modifier_rate': check_type.pricing_rate,
            }
        # pricing not found
        raise PricingBookingCheckTypeError(
            details={
                'check_type_group': check_type.group.slug,
                'check_type': check_type.slug,
                'reason': 'not-configured',
            }
        )
